/*
 * $Id: main.cpp,v 1.1 2004/07/16 13:40:42 tsuruoka Exp $
 */

#include <stdio.h>
#include <fstream>
#include <map>
#include <list>
#include <iostream>
#include <sstream>
#include "maxent.h"

using namespace std;

//const int NUM_TRAIN_SENTENCES = 99999999;
const int NUM_TRAIN_SENTENCES = 2000;

const int NUM_TEST_SENTENCES = 99999999;

struct Token
{
  string str;
  string pos;
  string prd;
  Token(string s, string p) : str(s), pos(p) {}
};

typedef vector<Token> Sentence;

extern ME_Model_Data me_model_data[];
  
ME_Sample mesample(const vector<Token> &vt, int i, const string & prepos);
void viterbi(vector<Token> & vt, const ME_Model & me);
string postag(const string & s, const ME_Model & me);

int train(const vector<Sentence> & vs, ME_Model & me)
{
  cerr << "extracting features...";
  int n = 0;
  for (vector<Sentence>::const_iterator i = vs.begin(); i != vs.end(); i++) {
    const Sentence & s = *i;
    for (int j = 0; j < s.size(); j++) {
      string prepos = "BOS";
      if (j > 0) prepos = s[j-1].pos;
      me.add_training_sample(mesample(s, j, prepos));
    }
  }
  cerr << "done" << endl;

  me.set_heldout(1000, 0);

  me.train(5, 1000, 0);
}

void deterministic(vector<Token> & vt, const ME_Model & me)
{
  for (int i = 0; i < vt.size(); i++) {
    string prepos = "BOS";
    if (i > 0) prepos = vt[i-1].prd;
    ME_Sample mes = mesample(vt, i, prepos);
    me.classify(mes);
    vt[i].prd = mes.label;
  }
}

void postagging(vector<Sentence> & vs, const ME_Model & me)
{
  int num_classes = me.num_classes();

  for (vector<Sentence>::iterator i = vs.begin(); i != vs.end(); i++) {
    Sentence & s = *i;
    viterbi(s, me);
    //    deterministic(s, me);
  }
}

void evaluate(const vector<Sentence> & vs)
{
  int ncorrect = 0;
  int ntotal = 0;
  for (vector<Sentence>::const_iterator i = vs.begin(); i != vs.end(); i++) {
    const Sentence & s = *i;
    for (int j = 0; j < s.size(); j++) {
      ntotal++;
      if (s[j].pos == s[j].prd) ncorrect++;
      cout << s[j].str << "\t" << s[j].pos << "\t" << s[j].prd << endl;
    }
  }
  fprintf(stderr, "%d / %d = %f\n", ncorrect, ntotal, (double)ncorrect / ntotal);
}

void read_pos(const string & filename, vector<Sentence> & vs, int num_sentences = -1)
{
  ifstream ifile(filename.c_str());

  string line;
  int n = 0;
  cerr << "reading " << filename;
  while (getline(ifile,line)) {
    istringstream is(line);
    string s;
    Sentence sentence;
    while (is >> s) {
      string::size_type i = s.find_last_of('/');
      string str = s.substr(0, i);
      string pos = s.substr(i+1);
      if (str == "-LRB-" && pos == "-LRB") { str = "("; pos = "("; }
      if (str == "-RRB-" && pos == "-RRB") { str = ")"; pos = ")"; }
      if (str == "-LSB-" && pos == "-LSB") { str = "["; pos = "["; }
      if (str == "-RSB-" && pos == "-RSB") { str = "]"; pos = "]"; }
      if (str == "-LCB-" && pos == "-LCB") { str = "{"; pos = "{"; }
      if (str == "-RCB-" && pos == "-RCB") { str = "}"; pos = "}"; }
      Token t(str, pos);
      sentence.push_back(t);
    }
    vs.push_back(sentence);
    if (vs.size() >= num_sentences) break;
    if (n++ % 10000 == 0) cerr << ".";
  }
  cerr << endl;

  ifile.close();
}


int main()
{
  vector<Sentence> trains, tests;

  read_pos("./wsj2-21.pos", trains, NUM_TRAIN_SENTENCES);

  ME_Model me;
  //  me.load_from_array(me_model_data);
  
  train(trains, me);
  me.save_to_file("./model");

  cerr << "evaluating wsj" << endl;
  read_pos("./wsj0.pos", tests, NUM_TEST_SENTENCES);
  postagging(tests, me);
  evaluate(tests);

}

/*
 * $Log: main.cpp,v $
 * Revision 1.1  2004/07/16 13:40:42  tsuruoka
 * init
 *
 */

